/* $NetBSD$ */

/*-
 * Copyright (c)2006, 2012 Takehiko NOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/*-
 * Copyright (c) 1992, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Original version ID:
 *     @(#)qsort.c	8.1 (Berkeley) 6/4/93
 *     NetBSD: qsort.c,v 1.21 2011/05/18 19:36:36 dsl Exp
 */

#include <sys/cdefs.h>
#if defined(LIBC_SCCS) && !defined(lint)
#ifdef __NetBSD__
__RCSID("$NetBSD$");
#endif
#endif /* LIBC_SCCS and not lint */

#include <assert.h>
#ifndef _DIAGASSERT
#define _DIAGASSERT(exp)	assert(exp)
#endif
#include <errno.h>
#include <stddef.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "tr24731_local.h"
#define fail(err, ...) \
do { \
	return __runtime_error(err, __func__, __VA_ARGS__); \
	/*NOTREACHED*/ \
} while (/*CONSTCOND*/0)
#include "tr24731_constraint.h"

#define min(a, b) (a) < (b) ? a : b

/*
 * Qsort routine from Bentley & McIlroy's "Engineering a Sort Function".
 */
#define swapcode(TYPE, parmi, parmj, n) { 		\
	rsize_t i = (n) / sizeof (TYPE); 		\
	TYPE *pi = (TYPE *)(void *)(parmi); 		\
	TYPE *pj = (TYPE *)(void *)(parmj); 		\
	do { 						\
		TYPE	t = *pi;			\
		*pi++ = *pj;				\
		*pj++ = t;				\
        } while (--i > 0);				\
}

#define SWAPINT(base0, size) \
	swaptype = ((char *)base0 - (char *)0) % sizeof(long) || \
        size % sizeof(long) ? 2 : size == sizeof(long) ? 0 : 1

static __inline void
swapfunc(char *a, char *b, rsize_t n, int swaptype)
{
	if (swaptype <= 1) 
		swapcode(long, a, b, n)
	else
		swapcode(char, a, b, n)
}

#define swap(a, b)						\
	if (swaptype == 0) {					\
		long t = *(long *)(void *)(a);			\
		*(long *)(void *)(a) = *(long *)(void *)(b);	\
		*(long *)(void *)(b) = t;			\
	} else							\
		swapfunc(a, b, size, swaptype)

#define vecswap(a, b, n) \
	if ((n) > 0) swapfunc((a), (b), (rsize_t)(n), swaptype)

static __inline char *
med3(char *a, char *b, char *c,
	int (*compar)(const void *, const void *, void *), void *context)
{
	return (compar(a, b, context) < 0)
	    ? ((compar(b, c, context) < 0)
		? b : ((compar(a, c, context) < 0) ? c : a))
            : ((compar(b, c, context) > 0)
		? b : ((compar(a, c, context) < 0) ? a : c));
}

errno_t
qsort_s(void *base, rsize_t nmemb, rsize_t size,
	int (*compar)(const void *, const void *, void *), void *context)
{
	char *base0;
	char *pa, *pb, *pc, *pd, *pm, *pn;
	rsize_t r;
	int d, swaptype;

	assert_rsize_max(nmemb);
	assert_rsize_max(size);
#ifdef COMPAT_WHIDBEY
	if (nmemb > 0)
		assert_null(base);
	assert_not_zero(size);
	assert_not_null(compar);
#else
	_DIAGASSERT(size > 0); /* XXX: FIXME */
	if (nmemb > 0) {
		assert_not_null(base);
		assert_not_null(compar);
	}
#endif
	if (nmemb < 1)
		return (errno_t)0;
	base0 = (char *)base;

loop:
	SWAPINT(base0, size);
	pm = base0 + (nmemb / 2) * size;
	if (nmemb > 7) {
		char *p = base0;
		pn = base0 + (nmemb - 1) * size;
		if (nmemb > 40) {
			r = (nmemb / 8) * size;
			p = med3(p, p + r, p + (2 * r), compar, context);
			pm = med3(pm - r, pm, pm + r, compar, context);
			pn = med3(pn - (2 * r), pn - r, pn, compar, context);
		}
		pm = med3(p, pm, pn, compar, context);
	}
	swap(base0, pm);
	pa = pb = base0 + size;
	pc = pd = base0 + (nmemb - 1) * size;
	for (;;) {
		while (pb <= pc &&
		    (d = compar(pb, base0, context)) <= 0) {
			if (d == 0) {
				swap(pa, pb);
				pa += size;
			}
			pb += size;
		}
		while (pb <= pc &&
		    (d = compar(pc, base0, context)) >= 0) {
			if (d == 0) {
				swap(pc, pd);
				pd -= size;
			}
			pc -= size;
		}
		if (pb > pc)
			break;
		swap(pb, pc);
		pb += size;
		pc -= size;
	}

	pn = base0 + (nmemb * size);
	r = min(pa - base0, pb - pa);
	vecswap(base0, pb - r, r);
	r = min(pd - pc , pn - pd - size);
	vecswap(pb, pn - r, r);
	r = pb - pa;
	if (r > size)
		qsort_s(base0, r / size, size, compar, context);
	r = pd - pc;
	if (r > size) {
		/* Iterate rather than recurse to save stack space */
		base0 = pn - r;
		nmemb = r / size;
		goto loop;
	}
	return (errno_t)0;
}
