/******************************************************************************

 Copyright (C) 2006 Frank Beythien, Michal Necasek

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 IN THE SOFTWARE.

******************************************************************************/


#define __STDC_WANT_LIB_EXT1__  1

#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include <search.h>
#include "safetest.h"

#if 1
#include <tr24731.h>
#endif

#define QSORT_SIZE  300
#define LIST_SIZE   200

int NumViolations = 0;      /* runtime-constraint violation counter */


/* Runtime-constraint handler for tests; doesn't abort program. */
void my_constraint_handler( const char *msg, void *ptr, errno_t error )
{
#ifdef PRINT_VIOLATIONS
    fprintf( stderr, "Runtime-constraint in %s", msg );
#endif
    ++NumViolations;
}

/*** Comparison routines ***/

int floatcmp_s( void const *_a, void const *_b, void *context )
{
    float   *a = (float *)_a;
    float   *b = (float *)_b;

    if( *a == *b ) {
        return 0;
    } else if( *a < *b ) {
        return -1;
    } else {
        return 1;
    }
}

/* compare routine with context parameter      */
int longintcmp_s( void const *_a, void const *_b, void *context )
{
    unsigned long   *a = (unsigned long *)_a;
    unsigned long   *b = (unsigned long *)_b;

    if( (*a) == (*b) ) {
        return 0;
    } else if( (*a) < (*b) ) {
        return -1;
    } else {
        return 1;
    }
}

/***************************************************************************/
/*  Test bsearch_s()                                                       */
/***************************************************************************/
void test_bsearch_s( void )
{
    int             violations = NumViolations;
    int             i;
    unsigned long   list[ LIST_SIZE ], cur;
    char            *context = "foo";


    /* prevent overflows in test data */
#if !( LIST_SIZE < 4000 )
    #error LIST_SIZE too large
#endif

    /* Create the search list */
    for( i = 0; i < LIST_SIZE; i++ ) {
        list[i] = i*i;
    }

    /* Test bsearch_s */
    for( i = 0; i < LIST_SIZE; i++ ) {
        cur = list[i];
        VERIFY( list[i] == i*i );
        VERIFY( bsearch_s( &cur, list, LIST_SIZE, sizeof( unsigned long ), longintcmp_s, context ) == &list[i] );
    }

    /* Test runtime-constraints */
#if 0
    VERIFY( bsearch_s( NULL, list, LIST_SIZE, sizeof( int ), longintcmp_s, context ) == NULL );
    VERIFY( NumViolations == ++violations );
#endif

    VERIFY( bsearch_s( &cur, NULL, LIST_SIZE, sizeof( int ), longintcmp_s, context ) == NULL );
    VERIFY( NumViolations == ++violations );

    VERIFY( bsearch_s( &cur, list, LIST_SIZE, sizeof( int ), NULL, context ) == NULL );
    VERIFY( NumViolations == ++violations );

    VERIFY( bsearch_s( &cur, NULL, LIST_SIZE, sizeof( int ), longintcmp_s, context ) == NULL );
    VERIFY( NumViolations == ++violations );

    VERIFY( bsearch_s( &cur, list, 0, sizeof( int ), longintcmp_s, context ) == NULL );
    VERIFY( NumViolations == violations );

#if RSIZE_MAX != SIZE_MAX
    VERIFY( bsearch_s( &cur, list, ~0, sizeof( int ), longintcmp_s, context ) == NULL );
    VERIFY( NumViolations == ++violations );

    VERIFY( bsearch_s( &cur, list, LIST_SIZE, ~0, longintcmp_s, context ) == NULL );
    VERIFY( NumViolations == ++violations );
#endif
}


/***************************************************************************/
/*  Test qsort_s()                                                         */
/***************************************************************************/
void test_qsort_s( void )
{
    int     violations = NumViolations;
    float   test[ QSORT_SIZE ];
    int     i;
    char    *context = "Bar";

    srand( 0 );
    for( i = 0; i < QSORT_SIZE; i++ ) {   /* Create test data */
        test[i] = rand( );
    }

    qsort_s( test, QSORT_SIZE, sizeof( float ), floatcmp_s, context ); /* Sort the data */

#if !( QSORT_SIZE - 1 > 0 )
    #error  QSORT_SIZE too small
#endif
    for( i = 0; i < QSORT_SIZE - 1; i++ ) {   /* Check the order */
        VERIFY( test[i] <= test[i+1] );
    }

    /* Test runtime-constraints */
    VERIFY( qsort_s( NULL, LIST_SIZE, sizeof( int ), floatcmp_s, context ) != 0 );
    VERIFY( NumViolations == ++violations );

    VERIFY( qsort_s( test, LIST_SIZE, sizeof( int ), NULL, context ) != 0 );
    VERIFY( NumViolations == ++violations );

    VERIFY( qsort_s( NULL, LIST_SIZE, sizeof( int ), floatcmp_s, context ) != 0 );
    VERIFY( NumViolations == ++violations );

    VERIFY( qsort_s( test, 0, sizeof( int ), floatcmp_s, context ) == 0 );
    VERIFY( NumViolations == violations );

#if RSIZE_MAX != SIZE_MAX
    VERIFY( qsort_s( test, ~0, sizeof( int ), floatcmp_s, context ) != 0 );
    VERIFY( NumViolations == ++violations );

    VERIFY( qsort_s( test, LIST_SIZE, ~0, floatcmp_s, context ) != 0 );
    VERIFY( NumViolations == ++violations );
#endif
}


int main( int argc, char *argv[] )
/********************************/
{
    set_constraint_handler_s( my_constraint_handler );

    test_bsearch_s();
    test_qsort_s();

    printf( "Tests completed (%s).\n", __FILE__ );
    return( 0 );
}
