/******************************************************************************

 Copyright (C) 2006 Frank Beythien, Michal Necasek

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 IN THE SOFTWARE.

******************************************************************************/


#define __STDC_WANT_LIB_EXT1__  1

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <time.h>
#include "safetest.h"


int         NumViolations = 0;          /* runtime-constraint violation counter */

/* Runtime-constraint handler for tests; doesn't abort program. */
void my_constraint_handler( const char *msg, void *ptr, errno_t error )
{
#ifdef PRINT_VIOLATIONS
    fprintf( stderr, "Runtime-constraint in %s", msg );
#endif
    ++NumViolations;
}


/***************************************************************************/
/*  Test Safer C Library time routines                                     */
/***************************************************************************/

void testtime_s( void )
/*********************/
{
    int                 violations = NumViolations;
    clock_t const       clocktime = clock();
    time_t              tt1;
    time_t              tt2;
    time_t              tt3;
    struct tm           tm1;
    struct tm           tm2;
    struct tm           gmt;
    char const * const  datestr = "Wed Aug 14 17:23:31 2002\n";
    char                buf[64];

    /*** Test various functions ***/

    tm1.tm_sec   = 31;
    tm1.tm_min   = 23;
    tm1.tm_hour  = 17;
    tm1.tm_mday  = 14;
    tm1.tm_mon   = 7;
    tm1.tm_year  = 102;
    tm1.tm_isdst = -1;
    tt1 = mktime( &tm1 );

    tm1.tm_sec++;
    tt2 = mktime( &tm1 );

    /* Test localtime_s() and asctime_s() */
    VERIFY( localtime_s( &tt1, &tm2 ) == &tm2 );

    VERIFY( asctime_s( buf, sizeof(buf), &tm2 ) == 0 );
    VERIFY( NumViolations == violations );
    VERIFY( strcmp( buf, datestr ) == 0 );

    VERIFY( localtime_s( NULL, &tm2 ) == NULL );
    VERIFY( NumViolations == ++violations );

    VERIFY( localtime_s( &tt1, NULL ) == NULL );
    VERIFY( NumViolations == ++violations );

    VERIFY( asctime_s( NULL, sizeof(buf), &tm2 ) != 0 );
    VERIFY( NumViolations == ++violations );

    buf[0] = 'X';
    VERIFY( asctime_s( buf, 25, &tm2 ) != 0 );
    VERIFY( NumViolations == ++violations );
    VERIFY( '\0' == buf[0] );

#if RSIZE_MAX != SIZE_MAX
    buf[0] = 'X';
    VERIFY( asctime_s( buf, ~0, &tm2 ) != 0 );
    VERIFY( NumViolations == ++violations );
    VERIFY( 'X' == buf[0] );
#endif

    tm2.tm_year = -1;
    buf[0] = 'X';
    VERIFY( asctime_s( buf, sizeof(buf), &tm2 ) != 0 );
    VERIFY( NumViolations == ++violations );
    VERIFY( '\0' == buf[0] );

    tm2.tm_year = 9999+1;
    buf[0] = 'X';
    VERIFY( asctime_s( buf, sizeof(buf), &tm2 ) != 0 );
    VERIFY( NumViolations == ++violations );
    VERIFY( '\0' == buf[0] );

    /*  ctime_s()                                       */
    VERIFY( ctime_s( buf, sizeof(buf), &tt1 ) == 0 );
    VERIFY( NumViolations == violations );
    VERIFY( strcmp( buf, datestr ) == 0 );

    VERIFY( ctime_s( NULL, sizeof(buf), &tt1 ) != 0 );
    VERIFY( NumViolations == ++violations );

    buf[0] = 'X';
    VERIFY( ctime_s( buf, 25, &tt1 ) != 0 );
    VERIFY( NumViolations == ++violations );
    VERIFY( '\0' == buf[0] );

#if RSIZE_MAX != SIZE_MAX
    buf[0] = 'X';
    VERIFY( ctime_s( buf, ~0, &tt1 ) != 0 );
    VERIFY( NumViolations == ++violations );
    VERIFY( 'X' == buf[0] );
#endif


    /* Test gmtime() and strtime() */
    VERIFY( gmtime_s( NULL, &tm2 ) == NULL);
    VERIFY( NumViolations == ++violations );

    VERIFY( gmtime_s( &tt2, NULL ) == NULL);
    VERIFY( NumViolations == ++violations );


    VERIFY( gmtime_s( &tt2 , &gmt ) != NULL );
    VERIFY( strftime( buf, sizeof( buf ), "%Y", &gmt ) == 4 );
    VERIFY( strcmp( buf, "2002" ) == 0 );
    VERIFY( strftime( buf, sizeof( buf ), "%m", &gmt ) == 2 );
    VERIFY( strcmp( buf, "08" ) == 0 );

    /* Make sure clock() isn't going backwards */
    VERIFY( clocktime <= clock() );

    /* Set TZ to UTC; this is so that mktime() doesn't use any offsets
     * and we can test the boundary values of time_t in the tests below.
     */
    setenv( "TZ", "GMT0", 1 );

    /* This time is the lowest to overflow */
    tm2.tm_sec   = 8;
    tm2.tm_min   = 14;
    tm2.tm_hour  = 3;
    tm2.tm_mday  = 19;
    tm2.tm_mon   = 0;
    tm2.tm_year  = 138;
    tm2.tm_isdst = -1;
    tt3 = mktime( &tm2 );
    if( tt3 != (time_t) -1 ) {
        VERIFY( gmtime_s( &tt3, &gmt ) == &gmt );
        VERIFY( tm2.tm_sec  == gmt.tm_sec );
        VERIFY( tm2.tm_min  == gmt.tm_min );
        VERIFY( tm2.tm_hour == gmt.tm_hour );
        VERIFY( tm2.tm_mday == gmt.tm_mday );
        VERIFY( tm2.tm_mon  == gmt.tm_mon );
        VERIFY( tm2.tm_year == gmt.tm_year );
    }

    /* This time is the greatest to work */
    tm2.tm_sec   -= 1;
    tt3 = mktime( &tm2 );
    VERIFY( tt3 != (time_t) -1 );
    VERIFY( gmtime_s( &tt3, &gmt ) == &gmt );
    VERIFY( tm2.tm_sec  == gmt.tm_sec );
    VERIFY( tm2.tm_min  == gmt.tm_min );
    VERIFY( tm2.tm_hour == gmt.tm_hour );
    VERIFY( tm2.tm_mday == gmt.tm_mday );
    VERIFY( tm2.tm_mon  == gmt.tm_mon );
    VERIFY( tm2.tm_year == gmt.tm_year );
}


/****
***** Program entry point.
****/

int main( int argc, char * const argv[] )
{
    set_constraint_handler_s( my_constraint_handler );

    testtime_s();

    printf( "Tests completed (%s).\n", __FILE__ );
    return( 0 );
}
